#define _c_APKOUTDB
/*********************************************************************/
/*                                                                   */
/* MODULE NAME: APKOUTDB.C                                           */
/*                                                                   */
/*                                                                   */
/* SYNOPSIS:  ACIF Sample output exit to perform MCF2 translation    */
/*            to allow the AFP Workbench Viewer to work with DBCS    */
/*            fonts.                                                 */
/*                                                                   */
/*                                                                   */
/* DESCRIPTION:  This exit searches for MFC2 records and converts    */
/*               them to NOP records.  It also searches for NOP      */
/*               records containing an MCF and changes them into     */
/*               an MCF.                                             */
/*                                                                   */
/*   FUNCTION = To use ACIF efficiently from the AFP Workbench       */ 
/*              Viewer with DBCS fonts, coded fonts must be used     */
/*              instead of code page/character set, so that the      */ 
/*              viewer does not need to look up each DBCS section    */ 
/*              as a separate font. ACIF generates a NOP containing  */ 
/*              the original MCF1 immediately following the new      */ 
/*              MCF2.                                                */ 
/*==>WARNING<== This exit's operation depends on an undocumented     */ 
/*==>WARNING<== ACIF function which places a copy of the original    */ 
/*==>WARNING<== MCF from the input file or PAGEDEF in a NOP record   */ 
/*==>WARNING<== following the generated MCF2. This is not a          */ 
/*==>WARNING<== guaranteed feature and may be changed without        */
/*==>WARNING<== notice.                                              */ 
/*                                                                   */
/* COPYRIGHT:                                                        */
/*  5765-140,5648-062 (C) COPYRIGHT IBM CORPORATION 1987,1993.       */
/*  All Rights Reserved                                              */
/*  Licensed Materials - Property of IBM                             */
/*                                                                   */
/*  US Government Users Restricted Rights - Use, duplication or      */
/*  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.*/
/*                                                                   */
/* NOTE: This program sample is provided on an as-is basis.          */
/*       The licensee of the PSF product is free to copy, revise,    */
/*       modify, and make derivative works of this program sample    */
/*       as they see fit.                                            */
/*                                                                   */
/* STATUS = RELEASE 1  LEVEL 0                                       */
/*                                                                   */
/*                                                                   */
/* EXTERNAL STATIC VARIABLES: None                                   */
/*                                                                   */
/*                                                                   */
/* NOTES:                                                            */
/*                                                                   */
/*                                                                   */
/* CHANGE ACTIVITY:                                                  */
/*                                                                   */
/* PROBNO  REL   DATE/AUTHOR DESCRIPTION                        FICHE*/
/* ------- ----- ----------- -----------                        -----*/
/*               8/94 T. Medearis  See below for description         */
/*                                                                   */
/*********************************************************************/         
/*                                                                   */   
/*                   <<<< IMPORTANT NOTE >>>>                        */
/*                                                                   */ 
/*  This sample ACIF Exit coding is provided solely as an example    */
/*  for study.  It has not been subjected to any formal IBM testing  */ 
/*  and is distributed on an "AS-IS" basis without any warranty      */ 
/*  either express or implied.  The use of this code, or any         */ 
/*  information or techniques implicit within it, is the sole        */ 
/*  responsibility of the customer.  Such use depends upon the       */ 
/*  customer's ability to evaluate the information in the context    */ 
/*  of their own particular operational environment.  This code      */ 
/*  has operated successfully in the computing environment at the    */ 
/*  Boulder Programming Center.        There is no guarantee that    */ 
/*  it will operate correctly elsewhere.  Customers attempting to    */ 
/*  adapt this code, or the techniques that it uses, to their own    */ 
/*  environment, do so entirely at their own risk.                   */ 
/*                                                                   */ 
/*********************************************************************/
/*  This program is a sample exit for ACIF (APKACIF,5648-062)        */
/*  that ...                                                         */
/*                                                                   */
/*  Method of Operation:                                             */
/*    IF EOF then set EOF flag                                       */
/*    ELSE                                                           */
/*     if an MCF2 structured field                                   */
/*       insert a NOP structured field                               */
/*     else                                                          */
/*       if a NOP followed by an MCF1 or MCF2 structured field then  */
/*         delete the NOP structured field.                          */
/*       else                                                        */
/*         do nothing                                                */
/*    END                                                            */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
#include "apkexits.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define sequence 0
#define MCF     0xD3B18A00     /* MCF structured field                  */
#define MCF2    0xD3AB8A00     /* MCF2 structured field                 */
#define NOP     0xD3EEEE00     /* Most of nop bytes                     */
#define FLAG    0x0000         /* Flag                                  */
#define SF      0x5A           /* structured field                      */

short  reclen;                 /* temp storage for record length        */
long   MCF2STG;                /* local var. used to move MCF2 constant */
long   MCFSTG;                 /* local var. used to move MCF constant  */
long   NOPSTG;                 /* local var. used to move NOP constant  */
short  seqstg;                 /* local var. used to move sequence const*/
char   flagstg;                /* local var. used to move flag conatant */
long   rc;                     /* return code                           */
long   rc2;                    /* intermediate return code              */
long   rc3;                    /* intermediate return code              */

long OUTEXIT( OUTEXIT_PARMS *exitstruc)
 {

     /***********************************************************/
     /*  Initialize ALL local variables.                        */
     /***********************************************************/
  rc = 0;
  rc2 = 0;
  rc3 = 0;
  MCF2STG = MCF2;
  MCFSTG = MCF;
  NOPSTG = NOP;
  seqstg = sequence;
  flagstg = FLAG;
  exitstruc->request = OUT_USE;      /* set no call back needed default  */
  if (exitstruc->eof==OUT_EOFLAG)
   {        
     exitstruc->request = OUT_EOFLAG;/* set no call back needed default  */
   }
  else
   {
     if ((rc=memcmp(&exitstruc->record[3],&MCF2STG,3))==0)
      {
            /***********************************************************/
            /*  After the first byte, shift the input record 8 bytes   */
            /*  to allow for the NOP structured field.                 */
            /***********************************************************/

        memmove(&reclen,                      /* load the local record length*/
                &exitstruc->record[1],2);     /* field.                      */
        reclen = reclen + 8;                  /* update the record length    */
        memmove(&exitstruc->record[9],        /* move all but first input    */
                &exitstruc->record[1],        /* to offset 9 in work area    */
                exitstruc->recordln-1);
        memmove(&exitstruc->record[1],        /* move the updated value into */
                &reclen,2);                   /* the record.                 */
        memmove(&exitstruc->record[3],        /* structured field id is NOP  */
                &NOPSTG,3);             
        memmove(&exitstruc->record[6],        /* set the flag byte           */
                &flagstg,1);          
        memmove(&exitstruc->record[7],        /* sequence value is irrelevant*/
                &seqstg,2);           
        exitstruc->recordln =                 /* update output record length */
                     exitstruc->recordln + 8; 
      }
     else
      {
            /***********************************************************/
            /*  If the record contains a NOP structured field          */
            /*  followed by an MCF1 or MCF2 structured field, then     */
            /*  delete the NOP structured field.                       */
            /***********************************************************/

        if ((((rc=memcmp(&exitstruc->record[3],&NOPSTG,3))==0) & \
            (((rc2=memcmp(&exitstruc->record[11],&MCFSTG,3))==0) || \
              (rc3=memcmp(&exitstruc->record[11],&MCF2STG,3))==0)))    
         {
           exitstruc->recordln =           /* reduce the output rec. length */
                   exitstruc->recordln - 8;
           memmove(&exitstruc->record[1],  /* eliminate the NOP field by    */
                   &exitstruc->record[9],  /* shifting the data at byte 9 to*/
                   (exitstruc->recordln)); /* byte 1.                       */
         }
        else
         {
            /***********************************************************/
            /*  The record was not an MCF2 or a NOP followed by an     */
            /*  MCF1 or MCF2, so do nothing.                           */
            /***********************************************************/
         }
      }  
   }    
  return( 0 );
 }
