#define _c_ASCIINP                                                              
/********************************************************************/          
/*  This program is a sample exit for ACIF (APKACIF,5648-062)       */          
/*  that accepts ASCII datastreams containing carriage return       */          
/*  and Formfeed characters and allows them to be formatted as      */          
/*  linedata using a PAGEDEF.                                       */          
/*                                                                  */          
/*                                                                  */          
/*  You must supply the ACIF control statement INPEXIT=ASCIINP      */          
/*  to enable this exit, and include the load module in a           */          
/*  dataset referenced in STEPLIB or elsewhere in the module        */          
/*  search order.  You must also specify the control statments      */          
/*  CC=YES and CCTYPE=Z.                                            */          
/*                                                                  */          
/*  Method of operation: for each input record, an additional       */          
/*  byte is inserted at the beginning of the record. Each byte      */          
/*  of the record is scanned. If an ASCII Formfeed character        */          
/*  (x'0C') is detected, all data up to the Formfeed is treated     */          
/*  as a record and returned to ACIF with the REQUEST byte set      */          
/*  to insert (x'02'). The Formfeed character is replaced by an     */          
/*  ANSI skip to channel 1 command (x'31'), and the remainder of    */          
/*  the record is inserted on the next call to this exit.           */          
/*                                                                  */          
/*  If a carriage return character (x'0D') appears in the first     */          
/*  or last byte of a record, the character is deleted and the      */          
/*  returned record length adjusted.  If a carriage return          */          
/*  appears elsewhere in the record, it is treated as Formfeed,     */          
/*  above, except the carriage return is replaced with an ASCII     */          
/*  print with no space command (x'2B').                            */          
/*                                                                  */          
/*  No other ASCII printer commands are recognized, but this        */          
/*  exit could easily be modified to account for vertical tab       */          
/*  (x'0B'), horizontal tab (x'09'), and backspace (x'08').         */          
/*                                                                  */          
/********************************************************************/          
#include "apkexits.h"                                                           
#include <string.h>                                                             
#include <stdio.h>                                                              
#include <stdlib.h>                                                             
                                                                                
#define MAX_LEN 32756          /* definition of length of storage area*/        
#define false  0                                                                
#define true 1                                                                  
#define skip2ch1 0x31          /* skip to channel 1 constant          */        
#define space    0x40          /* space - print the line and then     */        
                               /* carriage return and advance one line*/        
#define formfeed 0x0C          /* formfeed constant                   */        
#define crgret   0x0D          /* carriage return                     */        
#define prntline 0x2B          /* print the line, do not carriage     */        
                               /* return, do not linefeed.            */        
                                                                                
struct wrkarea                                                                  
     {                                                                          
        char    *stgrecptr;  /* pointer to 32k storage area in        */        
                             /* which portion of record that          */        
                             /* must be stored across multiple        */        
                             /* calls is stored.                      */        
        int     stgreclen;   /* length of data pointed to by          */        
                             /* stgrecptr.                            */        
        char    *stgendptr;  /* pointer to end of portion of          */        
                             /* saved on previous call + 1            */        
        char    res[4];      /* reserved                              */        
     };                                                                         
struct wrkarea *workarea;    /* pointer to the 16 byte work area      */        
int    remlength;            /* remaining length to be copied to the  */        
                             /* storage and used on the next invoca-  */        
                             /* tion of the exit.                     */        
int    recindex;             /* index into the input record.          */        
int    DONE;                 /* flag that indicates completion of loop*/        
                             /* processing                            */        
int    num=1;                /* number of arrays of data to allocate  */        
                             /*  on calloc invocation.                */        
char   *storTO;              /* pointer to where data is to be moved  */        
char   *inputTO;             /* pointer to where data is to be moved  */        
char   *FROM;                /* pointer to where data is to be moved  */        
                             /*   from.                               */        
char   *endptr;              /* pointer to end of record to be returned*/       
char   *recordptr;           /* pointer to record to be analyzed      */        
int    recordlen;            /* length of record to be analyzed       */        
int    returnlen;            /* length of record to be returned       */        
int    MOVESTG;              /* should data be moved from storage     */        
int    maxindex;             /* max byte offset into analyzed record  */        
                                                                                
long INPEXIT( INPEXIT_PARMS *exitstruc)                                         
{                                                                               
                                                                                
     /***********************************************************/              
     /*  Initialize ALL local variables.                        */              
     /***********************************************************/              
                                                                                
   remlength = 0;                                                               
   recindex = 0;                                                                
   storTO = 0;                                                                  
   inputTO = 0;                                                                 
   FROM = 0;                                                                    
   endptr = 0;                                                                  
   recordptr = 0;                                                               
   recordlen = 0;                                                               
   returnlen = 0;                                                               
   MOVESTG = false;                                                             
   maxindex = 0;                                                                
   exitstruc->request = INP_USE; /* set no call back needed default  */         
                                                                                
   workarea = (struct wrkarea *) exitstruc->work;                               
                                    /* set pointer to workarea     */           
                                                                                
   DONE = false;                    /* initialize DONE to false         */      
                                                                                
                                                                                
     /***********************************************************/              
     /*  If data from a previous call needs to be processed,    */              
     /*  then set the record pointer to point to the data in    */              
     /*  storage, and obtain the length of the record.          */              
     /***********************************************************/              
                                                                                
   if (workarea->stgreclen != 0)                                                
    {                                                                           
      recordptr = workarea->stgrecptr; /* work with record in storage   */      
      recordlen = workarea->stgreclen;                                          
    }                                                                           
   else                                                                         
    {                                                                           
      recordptr = exitstruc->record; /* work with record from input    */       
      recordlen = exitstruc->recordln;                                          
                                                                                
     }                                                                          
                                                                                
   endptr = recordptr + recordlen;       /* init endptr to equal the     */     
                                         /* recordpointer + length of rec*/     
   FROM = recordptr;                     /* if data is to be moved, the  */     
                                         /* source of the movement is the*/     
                                         /* beginning of the record.     */     
   MOVESTG = false;                      /* default is to not move data  */     
                                         /* from storage to exitstruc    */     
                                         /* area.                        */     
                                                                                
                                                                                
     /***********************************************************/              
     /*  Examine the first byte of the record for a carriage    */              
     /*  return or a formfeed.  If the first byte is either,    */              
     /*  then modifications to the byte will be necessary.      */              
     /*  If this call is examining data left over from a        */              
     /*  call then we assume that carriage returns should not   */              
     /*  advance the line.  Otherwise, we assume that a line    */              
     /*  advance is needed.  Form feeds always convert to a skip*/              
     /*  to channel 1.                                          */              
     /***********************************************************/              
                                                                                
   if ((recordptr[0] == crgret) || (recordptr[0] == formfeed))                  
    {                                    /* if first char of record is a */     
                                         /* carriage return or form feed */     
                                         /* then replace the first byte  */     
                                         /* accordingly                  */     
      inputTO = &exitstruc->record[0];   /* prepare to move data to      */     
                                         /* offset 0 of the input record */     
                                         /* area, because no space needs */     
                                         /* to be inserted.              */     
      if (recordptr[0] == crgret)                                               
       {                                                                        
         if (workarea->stgrecptr != 0)   /* if processing data from a    */     
          {                              /* previous call then assume    */     
            recordptr[0] = prntline;     /* no line feed.                */     
          }                                                                     
         else                            /* otherwise return and linefeed*/     
          {                                                                     
            recordptr[0] = space;                                               
          }                                                                     
       }                                                                        
      else                               /* recordptr[0] = formfeed     */      
       {                                                                        
         if (workarea->stgrecptr != 0)   /* for formfeeds, always code  */      
          {                              /* a page advance.             */      
            recordptr[0] = skip2ch1;                                            
          }                                                                     
         else                                                                   
          {                                                                     
            recordptr[0] = skip2ch1;                                            
          }                                                                     
       }                                                                        
      if (workarea->stgreclen == 0)      /* if working with input from  */      
       {                                 /* call, then no shift of data */      
                                         /* is necessary, because all   */      
                                         /* that need be done is change */      
                                         /* the carriage return or      */      
                                         /* formfeed.                   */      
       }                                                                        
      else                               /* otherwise working with data */      
       {                                 /* stored from earlier call.   */      
                                         /* set flag to indicate that   */      
                                         /* data must be moved from the */      
                                         /* workarea to the exitstruc   */      
         MOVESTG   = true;               /* data will be moved from the */      
                                         /* storage area to parm area   */      
       }                                                                        
    }                                                                           
   else                                                                         
                                                                                
     /***********************************************************/              
     /*  The first byte was not a carriage control or a formfeed*/              
     /*  This means that a space will be inserted into the      */              
     /*  record.  If the data is not from a previous call, then */              
     /*  shift the input data one byte.  Otherwise prepare      */              
     /*  for data to be moved from the workarea to 1 byte into  */              
     /*  exitstruc->record.                                     */              
     /***********************************************************/              
                                                                                
    {                                                                           
      inputTO = &exitstruc->record[1];    /* prepare to move data to  */        
                                          /* offset 1 in the input    */        
                                          /* area.  Since the first   */        
                                          /* byte of the record was   */        
                                          /* not a carriage return or */        
                                          /* formfeed, then a space   */        
                                          /* will be inserted in front*/        
                                          /* of the record.           */        
      if (workarea->stgreclen == 0)       /* examining input record   */        
       {                                  /* lead char was not a CR or*/        
                                          /* formfeed, so shift data  */        
                                          /* one byte and set the     */        
                                          /* byte at offest 0 to space*/        
         memmove(inputTO, FROM, recordlen);/* move the data from offset*/       
                                          /* 0 to offset 1.           */        
         exitstruc->record[0] = space;    /* set the first byte       */        
         recordlen++;                     /* increment record pointer */        
         inputTO = &exitstruc->record[0]; /* reset inputTO so that    */        
                                          /* recordlen is properly    */        
                                          /* determined.              */        
       }                                                                        
      else                                /* examining stored record  */        
       {                                  /* lead char was not a CR or*/        
                                          /* formfeed, so examine all */        
                                          /* of stored record and     */        
                                          /* eventually move data to  */        
                                          /* be returned to one byte  */        
                                          /* into input record area   */        
                                          /* of the record.           */        
         MOVESTG = true;                  /* data will be moved from  */        
                                          /* the storage area to parm */        
                                          /* area                     */        
       }                                                                        
    }                                                                           
                                                                                
   maxindex = recordlen - 1;              /* set max while loop index */        
                                                                                
     /***********************************************************/              
     /*  Search the record for a formfeed or a carriage control.*/              
     /*  This means that a space will be inserted into the      */              
     /*  record.  If the data is not from a previous call, then */              
     /*  shift the input data one byte.  Otherwise prepare      */              
     /*  for data to be moved from the workarea to 1 byte into  */              
     /*  exitstruc->record.                                     */              
     /***********************************************************/              
                                                                                
   while ((DONE == false)&(recindex <= maxindex))                               
    {                                                                           
      switch (recordptr[recindex])                                              
       {                                                                        
         case formfeed:             /* if current byte is a form feed   */      
                                                                                
            remlength = recordlen - /* length of remaining portion      */      
                       recindex;    /* of modified input record equals  */      
                                    /* the length of the record minus   */      
                                    /* the current index value.         */      
                                                                                
            endptr = &recordptr[recindex]; /* set ptr to end of         */      
                                          /* data to be returned and the*/      
                                          /* the beginning of data to be*/      
                                          /* stored.                    */      
                                                                                
     /***********************************************************/              
     /*  If not already obtained, obtain storage in which to    */              
     /*  store any data not yet examined.                       */              
     /***********************************************************/              
                                                                                
            if (workarea->stgreclen == 0)/* if no storage area has been */      
             {                           /*   obtained for that portion */      
               workarea->stgrecptr =/*   of any record that may need to */      
                       calloc( num, /*   be stored across calls, then   */      
                         MAX_LEN);  /*   get that storage now.          */      
               workarea->stgreclen =/* set length of record in storage  */      
                           remlength;                                           
                                                                                
     /***********************************************************/              
     /*  Move the remaining portion of modified input record    */              
     /*  to the 32k storage area.                               */              
     /***********************************************************/              
                                                                                
               storTO = &workarea->stgrecptr[0];                                
               FROM = &recordptr[recindex];                                     
               memmove(storTO, FROM, remlength);                                
                                                                                
             }                      /*                                  */      
            else                                                                
             {                          /* data already in storage.    */       
                                        /* just advance the pointer and*/       
                                        /* change the length           */       
               workarea->stgreclen =/* set length of record in storage  */      
                           remlength;                                           
               workarea->stgrecptr = &workarea->stgrecptr[recindex];            
             }                                                                  
            exitstruc->request = INP_INSERT; /* set insert request flag */      
                                    /* so that ACIF will call the exit  */      
                                    /* again.                           */      
            DONE = true;            /* set flag to exit loop.           */      
            break;                  /* end case FORM FEED               */      
                                                                                
                                                                                
     /***********************************************************/              
     /*  If a carriage return is found, then one scenario is    */              
     /*  for a carriage return at the end of the input record   */              
     /*  The other is for a carriage return occurring in the    */              
     /*  middle of the record.                                  */              
     /***********************************************************/              
                                                                                
         case crgret:               /* if current byte is a carriage ret*/      
            if (recindex == maxindex)                                           
                                                                                
             /***********************************************************/      
             /*  Handle a carriage return that occurs at the end of the */      
             /*  record.  The carriage return is discarded.             */      
             /***********************************************************/      
                                                                                
             {                                                                  
               endptr = &recordptr[recindex]; /* set ptr to end of      */      
                                          /* data to be returned and the*/      
                                          /* the beginning of data to be*/      
                                          /* stored.                    */      
                                                                                
               workarea->stgreclen = 0;   /* set the length of data left*/      
                                          /* to process in the work area*/      
                                          /* to zero, so that the work  */      
                                          /* area will be freed         */      
               DONE = true;               /* set flag to exit loop      */      
             }                                                                  
            else                                                                
             {                                                                  
                                                                                
             /***********************************************************/      
             /*  Handle a carriage return that occurs in the middle of  */      
             /*  the record.                                            */      
             /***********************************************************/      
                                                                                
               remlength = recordlen - /* length of remaining portion  */       
                       recindex;    /* of modified input record equals  */      
                                    /* the length of the record minus   */      
                                    /* the current index value.         */      
               endptr = &recordptr[recindex]; /* set ptr to end of      */      
                                          /* data to be returned and the*/      
                                          /* the beginning of data to be*/      
                                          /* stored.                    */      
                                                                                
             /***********************************************************/      
             /*  Get storage for that portion of the record including   */      
             /*  and following the carriage return.                     */      
             /***********************************************************/      
                                                                                
               if (workarea->stgrecptr == 0)/* if no storage area has been */   
                {                           /*   obtained for that portion */   
                  workarea->stgrecptr =/*   of any record that may need to */   
                          calloc( num, /*   be stored across calls, then   */   
                            MAX_LEN);  /*   get that storage now.          */   
                                                                                
                  workarea->stgreclen =/* set length of record in storage  */   
                              remlength;                                        
             /***********************************************************/      
             /*  Move the remaining portion of modified input record    */      
             /*  to the 32k storage area.                               */      
             /***********************************************************/      
                                                                                
                  storTO = workarea->stgrecptr;                                 
                  FROM = &recordptr[recindex];                                  
                  memmove(storTO, FROM, remlength);                             
                                                                                
                }                     /*                                  */    
               else                                                             
                {                          /* data already in storage.    */    
                                           /* just advance the pointer and*/    
                                           /* change the length           */    
                  workarea->stgreclen =/* set length of record in storage  */   
                              remlength;                                        
                  workarea->stgrecptr = &workarea->stgrecptr[recindex];         
                }                                                               
               DONE = true;         /* set flag to exit loop.           */      
               exitstruc->request = INP_INSERT; /* set insert request flag */   
                                    /* so that ACIF will call the exit  */      
                                    /* again.                           */      
             }                                                                  
            break;                  /* end case CARRIAGE RETURN         */      
         default:                                                               
            break;                                                              
                                                                                
       }                            /* end SWITCH statement             */      
      recindex++;                   /* increment the input record index */      
    }                               /* end WHILE loop                   */      
   returnlen = endptr - recordptr;  /* determine the length of the      */      
                                    /* record to be returned to the     */      
                                    /* caller.                          */      
                                                                                
             /***********************************************************/      
             /*  If data needs to be moved from the storage area to the */      
             /*  input area, then do it now.                            */      
             /***********************************************************/      
                                                                                
   if (MOVESTG == true)                                                         
    {                               /* move data in storage             */      
      memmove(inputTO, recordptr,   /* from the storecptr to 0 or 1     */      
              returnlen);           /* into the input record area for   */      
                                    /* a length up to but not including */      
                                    /* the formfeed or carriage control */      
      MOVESTG = false;              /* reset moveinput flag             */      
                                                                                
    }                                                                           
   exitstruc->recordln = inputTO -  /* the length of the data returned  */      
                                    /* depends on whether it was written*/      
                                    /* to offset 0 or 1.                */      
                         &exitstruc->record[0] +                                
                         returnlen;                                             
   if (workarea->stgrecptr != 0)    /* if there is no more data to be   */      
    {                               /* processed, then free the 32K     */      
      if (workarea->stgreclen <= 0)                                             
       {                                                                        
         free(workarea->stgrecptr); /* storage area.                    */      
         workarea->stgrecptr = 0;   /* set storage pointer to 0         */      
         workarea->stgreclen = 0;   /* set storage record length to 0   */      
       }                                                                        
    }                                                                           
   return(0);                                                                   
 }                                                                              
